# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=import-error

import pandas as pd

class MappingBase:
    key_name = None
    
    def __init__(self, mapping):
        if self.key_name not in mapping.columns:
            raise ValueError('Mapping must have column {}'.format(self.key_name))

        if 'pft' not in mapping.columns:
            raise ValueError('Mapping must have column pft')
        
        self._mapping = mapping

    @property
    def mapping(self): return self._mapping

class BiomePftMapping(MappingBase):
    key_name = 'biome'


class TaxaPftMapping(MappingBase):
    key_name = 'taxa'


class BiomePftMatrix(BiomePftMapping):
    def __init__(self, matrix):
        super().__init__(convert_matrix_to_mapping(matrix, self.key_name))


class TaxaPftMatrix(TaxaPftMapping):
    def __init__(self, matrix):
        super().__init__(convert_matrix_to_mapping(matrix, self.key_name))


class PftListBase:
    @staticmethod
    def _convert_list_to_mapping(list, key_name):
        list = clean_column_name(list, 0, key_name)
        list = clean_column_name(list, 1, 'pft')
        return list.explode('pft')

    @classmethod
    def read_csv(cls, filepath_or_buffer, **kwargs):
        key_name = cls.key_name # pylint: disable=no-member
        raw = pd.read_csv(filepath_or_buffer, dtype=str, header=None, **kwargs)
        df_list = raw.apply(lambda row: pd.Series([row.iloc[0], row.iloc[1:].dropna().to_list()], index=[key_name, 'pft']), axis='columns')
        return cls(df_list)


class BiomePftList(BiomePftMapping, PftListBase):
    def __init__(self, list):
        super().__init__(self._convert_list_to_mapping(list, self.key_name))


class TaxaPftList(TaxaPftMapping, PftListBase):
    def __init__(self, list):
        super().__init__(self._convert_list_to_mapping(list, self.key_name))


def clean_column_name(df, index, name):
    """Check that the column has the expected name (case insensitive)
    as a sanity check that the user provided the right data, 
    and return a new dataframe with the column renamed to the preferred casing.
    """
    df_name = str(df.columns[index])

    if df_name.lower() != name:
        raise ValueError('Column {} in the dataframe must be called "{}"'.format(index + 1, name))

    return df.rename(columns={ df_name: name })


def convert_matrix_to_mapping(matrix, key_name):
    matrix = clean_column_name(matrix, 0, key_name)

    # Convert the matrix into a list of relations between biomes/taxas and PFTs
    mapping = matrix.melt(id_vars=[key_name], var_name='pft', value_name='has_pft')
    return mapping[mapping.has_pft == 1].filter(items=[key_name, 'pft'])





# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=import-error

import pandas as pd
import itertools

def safe_int(s):
    return int(s) if s else 0

def dataframe_from_gspread_sheet(sheet, converter=str, first_value_column=0, **kwargs):
    def convert_column(column, value):
        if column < first_value_column:
            return value
        else:
            return converter(value)

    rows = sheet.get_all_values()
    columns = [c.strip() for c in rows[0]]
    data = [map(convert_column, row) for row in rows[1:]]
    return pd.DataFrame.from_records(data, columns=columns, **kwargs)



# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=missing-function-docstring missing-module-docstring import-error

import pytest
import pandas as pd

from brioche.tool import main

SEPS = pytest.mark.parametrize('sep', [',', ';'])

@SEPS
def test_pollen_counts_for_multiple_sites(tmp_path, sep):
    taxas = write_taxas(tmp_path, sep)
    biomes = write_biomes(tmp_path, sep)

    site1 = write_samples(tmp_path, 'site1', sep,
        ('depth', 'taxa1', 'taxa2', 'taxa3'),
        (10, 1, 4, 45))

    site2 = write_samples(tmp_path, 'site2', sep,
        ('depth', 'taxa1', 'taxa0'),
        (20, 1, 9))

    main(['--decimals=1', 
        '--separator', sep,
        '--save-percentages', 
        '--save-stabilized',
        '--taxas', taxas,
        '--biomes', biomes,
        site1, site2])

    assert read_csv(tmp_path / 'site1_percentages.csv') == expected_csv(sep,
        ('depth', 'taxa1', 'taxa2', 'taxa3'),
        (10, '2.0', '8.0', '90.0'))

    assert read_csv(tmp_path / 'site1_stabilized.csv') == expected_csv(sep,
        ('depth', 'taxa1', 'taxa2', 'taxa3'),
        (10, '1.2', '2.7', '9.5'))

    assert read_csv(tmp_path / 'site1_scores.csv') == expected_csv(sep,
        ('depth', 'biome1', 'biome2', 'biome3'),
        (10, '1.18', '2.68', '9.48'))

    assert read_csv(tmp_path / 'site1_biomes.csv') == expected_csv(sep,
        ('depth', 'Biome'),
        (10, 'biome3'))

    assert read_csv(tmp_path / 'site2_percentages.csv') == expected_csv(sep,
        ('depth', 'taxa1', 'taxa0'),
        (20, '10.0', '90.0'))

    assert read_csv(tmp_path / 'site2_stabilized.csv') == expected_csv(sep,
        ('depth', 'taxa1', 'taxa0'),
        (20, '3.1', '9.5'))

    assert read_csv(tmp_path / 'site2_scores.csv') == expected_csv(sep,
        ('depth', 'biome1', 'biome2', 'biome3'),
        (20, '12.58', '9.48', '9.48'))

    assert read_csv(tmp_path / 'site2_biomes.csv') == expected_csv(sep,
        ('depth', 'Biome'),
        (20, 'biome1'))

@SEPS
def test_pollen_percentages_with_high_default_threshold(tmp_path, sep):
    taxas = write_taxas(tmp_path, sep)
    biomes = write_biomes(tmp_path, sep)

    site1 = write_samples(tmp_path, 'site1', sep,
        ('depth', 'taxa1', 'taxa2', 'taxa3'),
        (10, 1, 9, 90),
        (20, 50, 40, 10))

    main(['--decimals=3', 
        '--separator', sep,
        '--default-threshold=10.5',
        '--save-stabilized',
        '--type=percentages',
        '--taxas', taxas,
        '--biomes', biomes,
        site1])

    assert read_csv(tmp_path / 'site1_stabilized.csv') == expected_csv(sep,
        ('depth', 'taxa1', 'taxa2', 'taxa3'),
        (10, '0.000', '0.000', '8.916'),
        (20, '6.285', '5.431', '0.000'))

    assert read_csv(tmp_path / 'site1_scores.csv') == expected_csv(sep,
        ('depth', 'biome1', 'biome2', 'biome3'),
        (10, '0.0000', '0.0000', '8.9158'),
        (20, '6.2848', '5.4308', '0.0000'))

    assert read_csv(tmp_path / 'site1_biomes.csv') == expected_csv(sep,
        ('depth', 'Biome'),
        (10, 'biome3'),
        (20, 'biome1'))


@SEPS
def test_stabilzed_pollen_samples(tmp_path, sep):
    taxas = write_taxas(tmp_path, sep)
    biomes = write_biomes(tmp_path, sep)

    site1 = write_samples(tmp_path, 'site1', sep,
        ('depth', 'taxa0', 'taxa2', 'taxa3'),
        (10, '1.10', '3.33', '0.00'))

    main(['--separator', sep,
        '--type=stabilized',
        '--taxas', taxas,
        '--biomes', biomes,
        site1])

    assert read_csv(tmp_path / 'site1_scores.csv') == expected_csv(sep,
        ('depth', 'biome1', 'biome2', 'biome3'),
        (10, '1.098', '4.428', '1.098'))

    assert read_csv(tmp_path / 'site1_biomes.csv') == expected_csv(sep,
        ('depth', 'Biome'),
        (10, 'biome2'))


# Simple taxa mapping: taxa1/2/3 maps to PFTs 1/2/3 respectively
# Also include taxa0 that maps to PFTs 1,2,3 to check that irregular CSV files work
def write_taxas(tmp_path, sep):
    taxa_file = tmp_path / 'taxas.csv'

    with open(taxa_file, 'wt') as f:
        f.write(f'taxa0{sep}1{sep}2{sep}3\n')
        for i in range(1,4):
            f.write(f'taxa{i}{sep}{i}\n')

    return str(taxa_file)

# Simple mapping: biome1/2/3 maps to PFTs 1/2/3 respectively
def write_biomes(tmp_path, sep):
    biome_file = tmp_path / 'biomes.csv'

    with open(biome_file, 'wt') as f:
        for i in range(1,4):
            f.write(f'biome{i}{sep}{i}\n')

    return str(biome_file)

def write_samples(tmp_path, site, sep, *rows):
    sample_file = tmp_path / f'{site}.csv'

    with open(sample_file, 'wt') as f:
        for row in rows:
            f.write(sep.join(map(str, row)))
            f.write('\n')

    return str(sample_file)

def read_csv(path):
    with open(path, 'rt') as f:
        return f.read()

def expected_csv(sep, *rows):
    return '\n'.join([sep.join(map(str, row)) for row in rows]) + '\n'


# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=missing-function-docstring missing-module-docstring import-error

import pandas as pd
import pytest

from brioche import \
    BiomePftMatrix, TaxaPftMatrix, \
    BiomePftList, TaxaPftList, \
    Biomization, PollenCounts, \
    StabilizedPollenSamples

def test_join_matrix_mappings_into_biome_matrix():
    # taxa1: maps only through PFT 1 to biome 2
    # taxa2: maps only through PFT 2 to biome 1
    # taxa3: maps only to PFT 3, which doesn't map to any biome
    # taxa4: maps through PFT 2/4 to biome1 and PFT 1/4 to biome 2

    taxa_pfts = TaxaPftMatrix(pd.DataFrame.from_records(
            columns=('taxa', 1, 2, 3, 4),
            data=[
                ('taxa1', 1, 0, 1, 0),
                ('taxa2', 0, 1, 1, 0),
                ('taxa3', 0, 0, 1, 0),
                ('taxa4', 1, 1, 1, 1)
                ]))

    biome_pfts = BiomePftMatrix(pd.DataFrame.from_records(
            columns=('biome', 1, 2, 4),
            data=[
                ('biome1', 0, 1, 1),
                ('biome2', 1, 0, 1)
                ]))

    biomization = Biomization(taxa_pfts, biome_pfts)

    result = biomization.taxa_biome_matrix.to_dict()

    assert result == dict(
        biome1=dict(taxa1=0, taxa2=1, taxa4=1),
        biome2=dict(taxa1=1, taxa2=0, taxa4=1)
    )

def test_join_list_mappings_into_biome_matrix():
    # taxa1: maps only through PFT 1 to biome 2
    # taxa2: maps only through PFT 2 to biome 1
    # taxa3: maps only to PFT 3, which doesn't map to any biome
    # taxa4: maps through PFT 2/4 to biome1 and PFT 1/4 to biome 2

    taxa_pfts = TaxaPftList(pd.DataFrame.from_records(
            columns=('taxa', 'pft'),
            data=[
                ('taxa1', ['1', '3']),
                ('taxa2', ['2', '3']),
                ('taxa3', ['3']),
                ('taxa4', ['1', '2', '3', '4'])
                ]))

    biome_pfts = BiomePftList(pd.DataFrame.from_records(
            columns=('biome', 'pft'),
            data=[
                ('biome1', ['2', '4']),
                ('biome2', ['1', '4'])
                ]))

    biomization = Biomization(taxa_pfts, biome_pfts)

    result = biomization.taxa_biome_matrix.to_dict()

    assert result == dict(
        biome1=dict(taxa1=0, taxa2=1, taxa4=1),
        biome2=dict(taxa1=1, taxa2=0, taxa4=1)
    )

def test_get_biome_affinity():
    taxa_pfts = TaxaPftMatrix(pd.DataFrame.from_records(
            columns=('taxa', 1, 2, 3),
            data=[
                ('taxa1', 1, 1, 0),
                ('taxa2', 0, 1, 1),
                ('taxa3', 0, 0, 1),
                ('taxa4', 0, 0, 1)
                ]))

    # Use a 1-1 mapping between PFTs and biomes to simplify things
    biome_pfts = BiomePftMatrix(pd.DataFrame.from_records(
            columns=('biome', 1, 2, 3),
            data=[
                ('biome1', 1, 0, 0),
                ('biome2', 0, 1, 0),
                ('biome3', 0, 0, 1)
                ]))

    samples = StabilizedPollenSamples(
        samples=pd.DataFrame.from_dict(
            orient='index',
            columns=('taxa1', 'taxa2', 'taxa3', 'taxa4'),
            data=dict(
                taxa1_only_maps_to_biome1_as_least_specific=[
                    5, 0, 0, 0
                ],
                taxa2_only_maps_to_biome2_as_least_specific=[
                    0, 10, 0, 0
                ],
                taxa1_and_2_maps_to_biome2_as_highest_affinity=[
                    5, 10, 0, 0
                ],
                all_taxas_maps_to_biome3_as_highest_affinity=[
                    1, 2, 3, 4
                ],
                no_samples_maps_to_na=[
                    0, 0, 0, 0
                ]
            )),
        decimals=0,
        site="test")

    biomization = Biomization(taxa_pfts, biome_pfts)
    affinity = biomization.get_biome_affinity(samples)
    assert affinity.site == "test"

    biome_results = affinity.biomes.to_dict()
    assert biome_results == dict(
        taxa1_only_maps_to_biome1_as_least_specific='biome1',
        taxa2_only_maps_to_biome2_as_least_specific='biome2',
        taxa1_and_2_maps_to_biome2_as_highest_affinity='biome2',
        all_taxas_maps_to_biome3_as_highest_affinity='biome3',
        no_samples_maps_to_na='N/A'
    )

    score_results = affinity.scores.to_dict(orient='index')

    assert score_results == dict(
        taxa1_only_maps_to_biome1_as_least_specific=dict(
            biome1=pytest.approx(4.9),
            biome2=pytest.approx(4.8),
            biome3=pytest.approx(0)
        ),
        taxa2_only_maps_to_biome2_as_least_specific=dict(
            biome1=pytest.approx(0),
            biome2=pytest.approx(9.8),
            biome3=pytest.approx(9.7)
        ),
        taxa1_and_2_maps_to_biome2_as_highest_affinity=dict(
            biome1=pytest.approx(4.9),
            biome2=pytest.approx(14.8),
            biome3=pytest.approx(9.7)
        ),
        all_taxas_maps_to_biome3_as_highest_affinity=dict(
            biome1=pytest.approx(0.9),
            biome2=pytest.approx(2.8),
            biome3=pytest.approx(8.7)
        ),
        no_samples_maps_to_na=dict(
            biome1=pytest.approx(0),
            biome2=pytest.approx(0),
            biome3=pytest.approx(0)
        )
    )

def test_order_of_taxas_in_samples_does_not_matter():
    taxa_pfts = TaxaPftMatrix(pd.DataFrame.from_records(
            columns=('taxa', 1),
            data=[
                ('taxa1', 1),
                ('taxa2', 1),
                ('taxa3', 0),
                ]))

    biome_pfts = BiomePftMatrix(pd.DataFrame.from_records(
            columns=('biome', 1),
            data=[
                ('biome1', 1),
                ]))

    samples = StabilizedPollenSamples(
        samples=pd.DataFrame.from_dict(
            orient='index',
            columns=('taxa3', 'taxa2', 'taxa1'),
            data=dict(row=[4, 2, 1])),
        decimals=0,
        site="test")

    biomization = Biomization(taxa_pfts, biome_pfts)
    affinity = biomization.get_biome_affinity(samples)

    result = affinity.scores.loc['row', 'biome1']
    assert result == pytest.approx(2.8)


def test_ignore_unmapped_taxas():
    taxa_pfts = TaxaPftMatrix(pd.DataFrame.from_records(
            columns=('taxa', 1),
            data=[
                ('taxa1', 1),
                ('taxa2', 1),
                ('taxa3', 0),
                ]))

    biome_pfts = BiomePftMatrix(pd.DataFrame.from_records(
            columns=('biome', 1),
            data=[
                ('biome1', 1),
                ]))

    samples = StabilizedPollenSamples(
        samples=pd.DataFrame.from_dict(
            orient='index',
            columns=('taxa1', 'taxa2', 'taxa4'),
            data=dict(row=[4, 2, 1])),
        decimals=0,
        site="test")

    biomization = Biomization(taxa_pfts, biome_pfts)
    affinity = biomization.get_biome_affinity(samples)

    result = affinity.scores.loc['row', 'biome1']
    assert result == pytest.approx(5.8)


def test_get_unmapped_taxas():
    taxa_pfts = TaxaPftMatrix(pd.DataFrame.from_records(
            columns=('taxa', 1),
            data=[
                ('taxa1', 1),
                ('taxa2', 1),
                ]))

    biome_pfts = BiomePftMatrix(pd.DataFrame.from_records(
            columns=('biome', 1),
            data=[
                ('biome1', 1),
                ]))

    samples1 = PollenCounts(
        samples=pd.DataFrame.from_dict(
            orient='index',
            columns=('taxa1', 'taxa2', 'taxa3', 'taxa4'),
            data=dict(row=[1, 2, 3, 4])))

    samples2 = PollenCounts(
        samples=pd.DataFrame.from_dict(
            orient='index',
            columns=('taxa1', 'taxa2', 'taxa4', 'taxa5'),
            data=dict(row=[1, 2, 4, 5])))

    biomization = Biomization(taxa_pfts, biome_pfts)
    result = biomization.get_unmapped_taxas(samples1, samples2)

    assert result == {'taxa3', 'taxa4', 'taxa5'}

@pytest.mark.parametrize(('number_of_taxas', 'expected_score'), [
    (1, 1.09),
    (9, 1.01),
    (10, 1.09),
    (11, 1.089),
    (99, 1.001),
    (100, 1.09),
    (191, 1.0809)
    ])
def test_tie_breaker_decrement_uses_enough_decimals_to_not_affect_affinity_score(number_of_taxas, expected_score):
    taxa_pfts = TaxaPftMatrix(pd.DataFrame.from_records(
            columns=('taxa', 1),
            data=[('taxa{}'.format(i), 1) for i in range(number_of_taxas)]))

    biome_pfts = BiomePftMatrix(pd.DataFrame.from_records(
            columns=('biome', 1),
            data=[
                ('biome1', 1),
                ]))

    samples = StabilizedPollenSamples(
        samples=pd.DataFrame.from_dict(
            orient='index',
            columns=('taxa0', ),
            data=dict(row=[1.1])),
        decimals=1)

    biomization = Biomization(taxa_pfts, biome_pfts)
    affinity = biomization.get_biome_affinity(samples)

    result = affinity.scores.loc['row', 'biome1']
    assert result == pytest.approx(expected_score)



# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=missing-function-docstring missing-module-docstring import-error

import pytest
import pandas as pd

from brioche import PollenCounts, PollenPercentages

def test_get_taxas():
    counts = PollenCounts(pd.DataFrame.from_records(
        columns=('Taxa A', 'Taxa B', 'Taxa C'),
        data=[(1, 0, 1)]))

    assert list(counts.taxas) == ['Taxa A', 'Taxa B', 'Taxa C']

def test_calculate_percentages_from_counts():
    counts = PollenCounts(pd.DataFrame.from_records(
        columns=('TaxaA', 'TaxaB', 'TaxaC'),
        data=[
            (5, 0, 15),
            (1, 2, 7),
            (0, 0, 0)
        ]),
        site="test")

    percentages = counts.get_percentages()
    result = percentages.samples.to_dict('records')

    assert result == [
        dict(TaxaA=25.0, TaxaB=0, TaxaC=75.0),
        dict(TaxaA=10.0, TaxaB=20.0, TaxaC=70.0),
        dict(TaxaA=0.0, TaxaB=0.0, TaxaC=0.0),
    ]

    assert counts.site == percentages.site


@pytest.mark.parametrize(('decimals', 'expected_value'), [
    (None, 100 / 6.0),
    (0, 17.0),
    (1, 16.7),
    (2, 16.67)])
def test_calculate_rounded_percentages_from_counts(decimals, expected_value):
    counts = PollenCounts(pd.DataFrame.from_records(
        columns=('TaxaA', 'TaxaB'),
        data=[
            (1, 5)
        ]))

    percentages = counts.get_percentages(decimals=decimals)
    result = percentages.samples.iat[0, 0]

    assert result == pytest.approx(expected_value)


@pytest.mark.parametrize(('decimals', 'expected_value'), [
    (None, 100 / 6.0),
    (0, 17.0),
    (1, 16.7),
    (2, 16.67)])
def test_get_rounded_percentages(decimals, expected_value):
    percentages = PollenPercentages(pd.DataFrame.from_records(
        columns=('TaxaA', ),
        data=[
            (100 / 6.0, )
        ]))

    rounded = percentages.get_percentages(decimals=decimals)
    result = rounded.samples.iat[0, 0]

    assert result == pytest.approx(expected_value)
    assert rounded.site == percentages.site


def test_get_percentages_without_rounding_returns_the_same_object():
    percentages = PollenPercentages(pd.DataFrame.from_records(
        columns=('TaxaA', ),
        data=[
            (100 / 6.0, )
        ]))

    result = percentages.get_percentages()

    assert result is percentages


@pytest.mark.parametrize(('percentage', 'threshold', 'decimals', 'expected'), [
    (25.5, 0.5, 2, 5.0),
    (1.1, 1.0, 2, 0.32),
    (1.1, 1.0, 3, 0.316),
    (1.0, 1.0, 5, 0.0),
    (0.1, 0.1, 5, 0.0),
    ])
def test_get_stabilized_using_default_threshold(percentage, threshold, decimals, expected):
    percentages = PollenPercentages(pd.DataFrame.from_records(
        columns=('TaxaA', ),
        data=[
            (percentage, )
        ]))

    stabilized = percentages.get_stabilized(default_threshold=threshold, decimals=decimals)
    result = stabilized.samples.iat[0, 0]

    assert result == pytest.approx(expected)



# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

"""Command line tool to run brioche on CSV files.
"""

import os
import argparse

from . import __version__
from .mappings import BiomePftList, TaxaPftList
from .samples import PollenCounts, PollenPercentages, StabilizedPollenSamples
from .biomization import Biomization

parser = argparse.ArgumentParser(description='Perform biome affinity analysis of pollen samples')
parser.add_argument('--version', action='version', version='%(prog)s {}'.format(__version__))
parser.add_argument('--separator', '-s', default=',', help='Column separator (default comma)', metavar='CHAR/REGEXP')
parser.add_argument('--decimals', type=int, choices=range(5), default=2, help='Decimals to use in stabilized sample values (default 2)')
parser.add_argument('--default-threshold', '-d', type=float, default=0.5, help='Default sample stabilization threshold (default 0.5)', metavar='THRESHOLD')
parser.add_argument('--type', choices=['counts', 'percentages', 'stabilized'], 
    default='counts', help='Type of values in the pollen sample files (default counts)')
parser.add_argument('--save-percentages', action='store_true', help='Save calculated sample percentages')
parser.add_argument('--save-stabilized', action='store_true', help='Save calculated stabilized sample values')
parser.add_argument('--taxas', '-t', required=True, help='Taxa to PFT mapping CSV file', metavar='TAXAS.CSV')
parser.add_argument('--biomes', '-b', required=True, help='Biome to PFT mapping CSV file', metavar='BIOMES.CSV')
parser.add_argument('samples', nargs='+', help='Pollen sample CSV files', metavar='SAMPLE.CSV')

def main(cli_args=None):
    args = parser.parse_args(cli_args)
    taxas = TaxaPftList.read_csv(args.taxas, sep=args.separator)
    biomes = BiomePftList.read_csv(args.biomes, sep=args.separator)
    samples = list(read_samples(args))

    biomization = Biomization(taxas, biomes)

    unmapped = biomization.get_unmapped_taxas(*samples)
    if unmapped:
        print('Warning: sample files contain taxas that are not mapped to any biome:')
        for t in unmapped:
            print(t)
        print()

    for sample in samples:
        print('Reading samples from:', sample.site)

        base = os.path.splitext(sample.site)[0]

        biomes_path = '{}_biomes.csv'.format(base)
        scores_path = '{}_scores.csv'.format(base)

        stabilized = sample.get_stabilized(default_threshold=args.default_threshold, decimals=args.decimals)

        if args.save_percentages:
            percentages_path = '{}_percentages.csv'.format(base)
            sample.get_percentages(args.decimals).to_csv(percentages_path, decimals=args.decimals, sep=args.separator)
            print('Wrote percentages to:', percentages_path)

        if args.save_stabilized:
            stabilized_path = '{}_stabilized.csv'.format(base)
            stabilized.to_csv(stabilized_path, sep=args.separator)
            print('Wrote stabilized to: ', stabilized_path)

        affinity = biomization.get_biome_affinity(stabilized)
        affinity.biomes_to_csv(biomes_path, sep=args.separator)
        affinity.scores_to_csv(scores_path, sep=args.separator)

        print('Wrote biomes to:     ', biomes_path)
        print('Wrote scores to:     ', scores_path)
        print()


def read_samples(args):
    for sample in args.samples:
        if args.type == 'counts':
            yield PollenCounts.read_csv(sample, site=sample, sep=args.separator)

        elif args.type == 'percentages':
            yield PollenPercentages.read_csv(sample, site=sample, sep=args.separator)

        elif args.type == 'stabilized':
            yield StabilizedPollenSamples.read_csv(sample, decimals=args.decimals, site=sample, sep=args.separator)



# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

"""Python/Pandas library to do biomization analyses of pollen samples
"""

__version__ = "0.2.1"

from .helpers import safe_int, dataframe_from_gspread_sheet

from .mappings import MappingBase, \
    BiomePftMapping, TaxaPftMapping, \
    BiomePftMatrix, TaxaPftMatrix, \
    BiomePftList, TaxaPftList

from .samples import PollenSamples, PollenCounts, PollenPercentages, StabilizedPollenSamples

from .biomization import Biomization



# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=import-error

import numpy as np
import pandas as pd

class PollenSamples:
    def __init__(self, samples, site=None):
        self._samples = samples
        self._site = site

    @classmethod
    def read_csv(cls, filepath_or_buffer, site=None, **kwargs):
        return PollenSamples._read_csv(cls, filepath_or_buffer, site=site, **kwargs)
    
    @staticmethod
    def _read_csv(constructor, filepath_or_buffer, site=None, **kwargs):
        df = pd.read_csv(filepath_or_buffer, index_col=0, header=0, **kwargs)
        return constructor(df, site=site)

    @property
    def samples(self): return self._samples

    @property
    def taxas(self): return self._samples.columns
        
    @property
    def site(self): return self._site

    def get_percentages(self, decimals=None): 
        raise NotImplementedError()

    def get_stabilized(self, default_threshold=0.0, decimals=2):
        percentages = self.get_percentages()

        # TODO: support per-taxa thresholds
        stabilized = (percentages.samples
                        .sub(default_threshold)
                        .clip(lower=0)
                        .apply(np.sqrt)
                        .round(decimals))

        return StabilizedPollenSamples(stabilized, site=self.site, decimals=decimals)

    def to_csv(self, path_or_buf, decimals=2, **kwargs):
        self.samples.to_csv(path_or_buf, float_format='%.{}f'.format(decimals), **kwargs)


class PollenCounts(PollenSamples):
    def get_percentages(self, decimals=None): 
        sums = self.samples.sum(axis=1)
        percentages = self.samples.apply(lambda column: column * 100 / sums).fillna(0.0)

        if decimals is not None:
            percentages = percentages.round(decimals)

        return PollenPercentages(percentages, self.site)


class PollenPercentages(PollenSamples):
    def get_percentages(self, decimals=None):
        if decimals is None:
            return self
        else:
            return PollenPercentages(self.samples.round(decimals), self.site)


class StabilizedPollenSamples(PollenSamples):
    def __init__(self, samples, decimals, site=None):
        super().__init__(samples, site=site)
        self._decimals = decimals

    @property
    def decimals(self): return self._decimals

    def get_stabilized(self, default_threshold=0.0, decimals=2):
        # TODO: round if decimals are fewer than this is set up to use
        return self

    @classmethod
    def read_csv(cls, filepath_or_buffer, decimals, site=None, **kwargs):
        return PollenSamples._read_csv(lambda samples, site=None: cls(samples, decimals, site=site), filepath_or_buffer, site=site, **kwargs)

    def to_csv(self, path_or_buf, decimals=None, **kwargs):
        if decimals is None:
            decimals = self.decimals

        super().to_csv(path_or_buf, decimals=decimals, **kwargs)


# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=missing-function-docstring missing-module-docstring import-error

import pytest
import pandas as pd

from brioche import BiomePftMatrix, TaxaPftMatrix, BiomePftList, TaxaPftList

def test_create_pft_biome_matrix():
    pft_biomes = BiomePftMatrix(pd.DataFrame.from_records(
        columns=('Biome', 1, 2, 3),
        data=[
            ('Biome A', 0, 1, 0),
            ('Biome B', 1, 1, 0)
            ]))

    result = sorted(pft_biomes.mapping.to_dict('records'), key=lambda d: (d['pft'], d['biome']))

    assert result == [
        dict(pft=1, biome='Biome B'),
        dict(pft=2, biome='Biome A'),
        dict(pft=2, biome='Biome B'),
    ]


def test_create_pft_taxa_matrix():
    pft_taxas = TaxaPftMatrix(pd.DataFrame.from_records(
        columns=('Taxa', 1, 2, 3),
        data=[
            ('Taxa A', 1, 0, 1),
            ('Taxa B', 0, 1, 1)
            ]))
            
    result = sorted(pft_taxas.mapping.to_dict('records'), key=lambda d: (d['pft'], d['taxa']))

    assert result == [
        dict(pft=1, taxa='Taxa A'),
        dict(pft=2, taxa='Taxa B'),
        dict(pft=3, taxa='Taxa A'),
        dict(pft=3, taxa='Taxa B'),
    ]


@pytest.mark.parametrize('mapping_class', [BiomePftMatrix, TaxaPftMatrix])
def test_matrix_must_contain_key_column(mapping_class):
    data = pd.DataFrame.from_records(
        columns=('foo', 1, 2),
        data=[('bar', 0, 1)])

    with pytest.raises(ValueError):
        mapping_class(data)


def test_create_biome_pft_list():
    biome_pfts = BiomePftList(pd.DataFrame.from_records(
        columns=('Biome', 'PFT'),
        data=[
            ('Biome A', [2, 3]),
            ('Biome B', [1, 2])
            ]))

    result = sorted(biome_pfts.mapping.to_dict('records'), key=lambda d: (d['pft'], d['biome']))

    assert result == [
        dict(pft=1, biome='Biome B'),
        dict(pft=2, biome='Biome A'),
        dict(pft=2, biome='Biome B'),
        dict(pft=3, biome='Biome A')
    ]


def test_create_taxa_pft_list():
    taxa_pfts = TaxaPftList(pd.DataFrame.from_records(
        columns=('Taxa', 'PFT'),
        data=[
            ('Taxa A', [3, 1]),
            ('Taxa B', [4, 1])
            ]))

    result = sorted(taxa_pfts.mapping.to_dict('records'), key=lambda d: (d['pft'], d['taxa']))

    assert result == [
        dict(pft=1, taxa='Taxa A'),
        dict(pft=1, taxa='Taxa B'),
        dict(pft=3, taxa='Taxa A'),
        dict(pft=4, taxa='Taxa B'),
    ]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['brioche']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['pandas ~= 1.1', 'gspread ~= 3.6', 'numpy ~= 1.1']

entry_points = \
{'console_scripts': ['brioche = brioche.tool:main']}

setup(name='brioche',
      version='0.2.1',
      description='Python/Pandas library to do biomization analyses of pollen samples',
      author='Peter Liljenberg',
      author_email='peter.liljenberg@gmail.com',
      url='https://github.com/petli/brioche',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='~=3.6',
     )



# Copyright 2020 Peter Liljenberg <peter.liljenberg@gmail.com>
# Open source under the MIT license (see LICENSE)

# pylint: disable=import-error

import math
import pandas as pd

class Biomization:
    def __init__(self, pft_taxas, pft_biomes):
        # Join the mappings on PFTs to get relationships between taxas and biomes
        self._taxa_biome_mapping = pft_biomes.mapping.merge(
            pft_taxas.mapping, on='pft', how='outer', sort=False)

        # Turn the relationships back into a matrix indexed by taxa with one column per biome
        self._taxa_biome_matrix = self._taxa_biome_mapping.pivot_table(
            index='taxa', columns='biome', values='pft',
            aggfunc=lambda v: 1,
            fill_value=0)

        # Count the mapped taxas per biome to calculate specificity adjustments below
        self._taxas_per_biome = self._taxa_biome_matrix.sum(axis='index')
        self._specificity_decimals = int(math.log10(self._taxas_per_biome.max())) + 1
        self._unmapped_taxas = set()

    @property
    def taxa_biome_mapping(self): return self._taxa_biome_mapping

    @property
    def taxa_biome_matrix(self): return self._taxa_biome_matrix

    def get_unmapped_taxas(self, *sites):
        return {
            taxa
            for site in sites
            for taxa in site.samples
            if taxa not in self._taxa_biome_matrix.index
        }

    def get_biome_affinity(self, stabilized_samples):
        samples = stabilized_samples.samples

        # Build up a data frame for the affinity scores by starting
        # with an empty one using the same index as the samples
        # (which is probably sample depth), then calculate an 
        # affinity score for each biome.
        scores = pd.DataFrame(index=samples.index)

        for biome in self._taxa_biome_matrix:
            # A biome with fewer mapped taxas has higher precedence than a more specific biome,
            # which we express as a fractional number below the number of decimals used in the
            # samples.
            decimals = stabilized_samples.decimals + self._specificity_decimals
            specificity = self._taxas_per_biome[biome] * 10 ** -decimals

            # Calculate an affinity score for the biome by multiplying the stablized sample values
            # with the 1s and 0s from the mapping, i.e. filtering out taxas that aren't mapped
            # and summing the remaining ones.  Deduct the specificity to give biomes with
            # more mapped taxas a slightly lower score than the ones with fewer, to break any ties
            # between them.
            biome_scores = (samples.mul(self._taxa_biome_matrix[biome], axis='columns')
                                .sum(axis='columns')
                                .sub(specificity)
                                .clip(lower=0))

            # Add the resulting series into the score data frame
            scores[biome] = biome_scores

        # For each row find the column with the highest value, i.e. the biome with the highest affinity score
        biomes = scores.idxmax(axis='columns')
        biomes.name = 'Biome'

        # Remove rows where there are no values
        score_sums = scores.sum(axis=1)
        biomes[score_sums == 0] = 'N/A'

        return BiomeAffinity(biomes, scores, stabilized_samples.site, decimals)


class BiomeAffinity:
    def __init__(self, biomes, scores, site, decimals):
        self._biomes = biomes
        self._scores = scores
        self._site = site
        self._decimals = decimals

    @property
    def biomes(self): return self._biomes

    @property
    def scores(self): return self._scores

    @property
    def site(self): return self._site

    def biomes_to_csv(self, path_or_buf, **kwargs):
        self.biomes.to_csv(path_or_buf, **kwargs)

    def scores_to_csv(self, path_or_buf, **kwargs):
        self.scores.to_csv(path_or_buf, float_format='%.{}f'.format(self._decimals), **kwargs)
